"""
w2lapp.modify.py: modify an entry

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: modify.py,v 1.68 2005/06/15 16:40:52 michael Exp $
"""

import ldap,ldif,ldap.schema, \
       pyweblib.forms, \
       w2lapp.core,w2lapp.cnf,w2lapp.gui,w2lapp.addmodifyform
       
from w2lapp.schema.syntaxes import syntax_registry,LDAPSyntaxValueError

#from ldap.modlist import modifyModlist
from ldaputil.modlist2 import modifyModlist

def GetEntryfromInputForm(form,ls,dn,sub_schema):

  invalid_attrs = []

  e = ldap.schema.Entry(sub_schema,dn.encode(ls.charset),{})
  try:
    in_ldif = form.field['in_ldif'].getLDIFRecords()
  except ValueError,e:
    raise w2lapp.core.ErrorExit('LDIF parsing error: %s' % (str(e)))
  else:
    if in_ldif:
      e.update(in_ldif[0][1])

  # Get all the attribute types
  in_attrtype_list = [
    a.encode('ascii')
    for a in form.getInputValue('in_attrtype',[])
  ]
  in_attrtype_list.extend([
    a.encode('ascii')
    for a in form.getInputValue('in_binattrtype',[])
  ])
  # Grab the Unicode input strings and convert to LDAP charset
  in_value_list = [
    a.encode(ls.charset)
    for a in form.getInputValue('in_value',[])
  ]
  # Grab the file-based input and leave as is
  in_value_list.extend([
    a for a in form.getInputValue('in_binvalue',[])
  ])
  
  if len(in_attrtype_list)!=len(in_value_list):
    raise w2lapp.core.ErrorExit('Different count of attribute types and values.')

  for i in range(len(in_attrtype_list)):
    attr_type = in_attrtype_list[i]
    syntax_class = syntax_registry.syntaxClass(sub_schema,attr_type)
    attr_instance = syntax_class(attr_type,None)
    try:
      attr_instance.setAttrValue(attr_instance.sanitizeInput(in_value_list[i]))
    except LDAPSyntaxValueError:
      invalid_attrs.append(attr_type)
    else:
      if attr_instance.attrValue:
        if not e.has_key(attr_type):
          e[attr_type] = []
        e[attr_type].append(attr_instance.attrValue)

  return e,invalid_attrs # GetEntryfromInputForm()


def ModlistTable(schema,modlist):
  """
  Return a string containing a HTML table showing modifiy operations
  and attr type/value pairs
  """
  s = []
  s.append('<table summary="Modify list">')
  for mod_op,mod_type,mod_val in modlist:
    if mod_val is None:
      mod_val_str = ''
    elif w2lapp.schema.no_humanreadable_attr(schema,mod_type):
      mod_val_str = '%s bytes of binary data' % (
        ' + '.join(
          [ str(len(x)) for x in mod_val ]
        )
      )
    else:
      mod_val_str = '<br>'.join([
        pyweblib.forms.escapeHTML(repr(v))
        for v in mod_val
      ])
    s.append('<tr><td>%s</td><td>%s</td><td>%s</td></tr>' % (
        ldif.MOD_OP_STR[mod_op],pyweblib.forms.escapeHTML(mod_type),mod_val_str
      )
    )
  s.append('</table>')
  return '\n'.join(s) # ModlistTable()


##############################################################################
# Modify existing entry
##############################################################################

def w2l_Modify(sid,outf,command,form,ls,dn):

  sub_schema = ls.retrieveSubSchema(dn,w2lapp.cnf.GetParam(ls,'_schema',None))

  new_entry,invalid_attrs = GetEntryfromInputForm(form,ls,dn,sub_schema)

  if invalid_attrs:
    w2lapp.addmodifyform.w2l_ModifyForm(
      sid,outf,'modifyform',form,ls,dn,
      Msg='Syntax check failed for the following attributes: %s' % (
        ', '.join(invalid_attrs)
      ),
      entry=new_entry,
      skip_oc_input=1
    )
    return

  in_oldattrtypes = {}
  for a in form.getInputValue('in_oldattrtypes',[]):
    in_oldattrtypes[a.encode('ascii')] = None

  # Read all attributes of entry which were viewable to the user
  # in the input form before
  result_dnlist = ls.readEntry(dn,in_oldattrtypes.keys(),no_cache=1)
  if not result_dnlist:
    raise w2lapp.core.ErrorExit("""
      No search result when reading entry <strong>%s</strong>.""" % (
        w2lapp.gui.DisplayDN(sid,form,ls,dn)
      )
    )

  old_entry = w2lapp.addmodifyform.ReadOldEntry(ls,dn,sub_schema)

  # All attributes currently read which were not viewable before
  # must be ignored to avoid problems with different access rights
  ignore_attr_types = [
    a
    for a in old_entry.keys()
    if not in_oldattrtypes.has_key(a)
  ]
  ignore_attr_types.extend([
    a
    for a in new_entry.keys()
    if w2lapp.schema.no_humanreadable_attr(sub_schema,a)
  ])
  # Create modlist containing deltas
  modlist = modifyModlist(
    sub_schema,
    old_entry,new_entry,
    ignore_attr_types=ignore_attr_types,
    ignore_oldexistent=0,
  )
  # Binary values are always replaced
  for a in new_entry.keys():
    if new_entry[a] and w2lapp.schema.no_humanreadable_attr(sub_schema,a):
      modlist.append((ldap.MOD_REPLACE,a,new_entry[a]))

  if modlist:
    # Send modify-list to host
    try:
      ls.modifyEntry(dn,modlist)
    except (
      ldap.CONSTRAINT_VIOLATION,
      ldap.INVALID_SYNTAX,
      ldap.OBJECT_CLASS_VIOLATION,
      ldap.OTHER,
      ldap.OTHER,
      ldap.TYPE_OR_VALUE_EXISTS,
      ldap.UNDEFINED_TYPE,
    ),e:
      w2lapp.addmodifyform.w2l_ModifyForm(
        sid,outf,'modifyform',form,ls,dn,
        Msg=w2lapp.gui.LDAPError2ErrMsg(e),
        entry=new_entry,
        skip_oc_input=1
      )
      return
    else:
      # delete all cache entries referencing to old DN
      UserMsg = 'Modified entry %s with your input data:</p>\n%s' % (
        w2lapp.gui.DisplayDN(sid,form,ls,dn),
        ModlistTable(sub_schema,modlist).encode(form.accept_charset)
      )
  else:
    UserMsg = 'No attributes modified of entry %s.' % (w2lapp.gui.DisplayDN(sid,form,ls,dn))

  # Output comes here
  w2lapp.gui.SimpleMessage(
    sid,outf,form,ls,dn,
    'Modify result',
    UserMsg,
    main_menu_list=w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=w2lapp.gui.ContextMenuSingleEntry(sid,form,ls,dn)
  )

