"""
w2lapp.schema.syntaxes: classes for known attribute types

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: syntaxes.py,v 1.79 2005/06/08 12:13:47 michael Exp $
"""

import re,imghdr,sndhdr,urllib,time,utctime,ldap,ldapurl,pyweblib.forms, \
       w2lapp.viewer,w2lapp.form,w2lapp.gui,mspki.util,w2lapp.cnf

from types import StringType,UnicodeType,IntType,LongType,FloatType,ClassType

from w2lapp.core import utf2display

from ldaputil.base import is_dn

try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO

try:
  from ldapoidreg import oid as oid_desc_reg
except ImportError:
  oid_desc_reg = {}


class SyntaxRegistry:

  def __init__(self):
    self.oid2syntax = ldap.cidict.cidict()
    self.at2syntax = ldap.cidict.cidict()

  def registerSyntaxClass(self,c):
    if type(c) is ClassType and hasattr(c,'oid'):
      self.oid2syntax[c.oid] = c

  def registerAttrType(self,syntax_oid,attrTypes):
    for a in attrTypes:
      self.at2syntax[a.strip()] = syntax_oid

  def syntaxClass(self,schema,attrtype_nameoroid):
    if attrtype_nameoroid is None:
      return LDAPSyntax
    attrtype_oid = schema.getoid(ldap.schema.AttributeType,attrtype_nameoroid)
    if self.at2syntax.has_key(attrtype_oid):
      syntax_class = self.oid2syntax[self.at2syntax[attrtype_oid]]
    else:
      attrtype_se = schema.get_inheritedobj(ldap.schema.AttributeType,attrtype_oid,['syntax'])
      if attrtype_se is None or attrtype_se.syntax is None:
        syntax_class = LDAPSyntax
      else:
        syntax_class = self.oid2syntax.get(attrtype_se.syntax,LDAPSyntax)
    return syntax_class 

  def attrInstance(self,schema,attrType,attrValue):
    syntax_class = self.syntaxClass(schema,attrType)
    attr_instance = syntax_class(attrType,attrValue)
    return attr_instance
    

url_pattern  = r'^(ftp|http|https|news|snews|ldap|ldaps|mailto):(|//)[^ ]*'
url_regex  = re.compile(url_pattern)
labeleduri_regex = re.compile(url_pattern+r' .*')
timestamp_pattern = r'^([0-9]){12,14}(.[0-9]+)*Z$'
timestamp_regex  = re.compile(timestamp_pattern)
mail_pattern = r'^[\w@.=/_ -]+@[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)*$'
mail_regex = re.compile(mail_pattern)


####################################################################
# Classes of known syntaxes
####################################################################


class LDAPSyntaxValueError(ValueError):
  pass


class LDAPSyntaxRegexNoMatch(LDAPSyntaxValueError):
  pass


class LDAPSyntax:
  oid=''
  desc='Any LDAP syntax'
  maxLen = w2lapp.cnf.misc.input_maxfieldlen
  maxValues = w2lapp.cnf.misc.input_maxattrs
  mimeType = 'application/octet-stream'
  fileExt = 'bin'
  editable = 1
  charset = 'utf-8'

  def __init__(self,attrType,attrValue):
    self.attrType = attrType
    self.attrValue = attrValue

  def setAttrValue(self,attrValue):
    self.validate(attrValue)
    self.attrValue = attrValue

  def sanitizeInput(self,attrValue):
    return attrValue

  def _regexValidate(self,attrValue):
    if hasattr(self,'reObj') and self.reObj.match(attrValue) is None:
      raise LDAPSyntaxRegexNoMatch, \
        "Class %s: %s does not match pattern %s." % (
          self.__class__.__name__,repr(attrValue[0:]),repr(self.reObj.pattern)
        )

  def _validate(self,attrValue):
    return 1

  def validate(self,attrValue):
    if attrValue:
      if not self._validate(attrValue):
        raise LDAPSyntaxValueError, \
          "Class %s: %s does not comply to syntax." % (
            self.__class__.__name__,repr(attrValue)
          )
      self._regexValidate(attrValue)

  def formField(self,form):
    input_field = pyweblib.forms.Input(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      self.maxLen,self.maxValues,None,default=None,size=50
    )
    input_field.charset = form.accept_charset
    input_field.setDefault(unicode(self.attrValue or '',self.charset))
    return input_field

  def getMimeType(self,form):
    return self.mimeType

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    if ldapurl.isLDAPUrl(self.attrValue):
      return LDAPUrl(self.attrType,self.attrValue).displayValue(sid,form,ls,dn,valueindex,commandbutton)
    rm = url_regex.search(self.attrValue)
    if rm!=None:
      rm = labeleduri_regex.match(self.attrValue)
      if rm!=None:
        url,label = self.attrValue.split(' ',1)
        return '<a href="%s/urlredirect?%s">%s</a>' % (form.script_name,url,utf2display(form.accept_charset,label))
      else:
        return '<a href="%s/urlredirect?%s">%s</a>' % (form.script_name,self.attrValue,self.attrValue)
    # Check for timestamp
    rm = timestamp_regex.match(self.attrValue)
    if rm!=None:
      return utctime.strftimeiso8601(utctime.strptime(self.attrValue))
    # No special format detected
    return utf2display(form.accept_charset,self.attrValue)


class Binary(LDAPSyntax):
  oid = '1.3.6.1.4.1.1466.115.121.1.5'
  desc = 'Binary'
  editable = 0

  def formField(self,form):
    f = pyweblib.forms.File(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      self.maxLen,self.maxValues,None,default=self.attrValue,size=50
    )
    f.mimeType = self.mimeType
    return f

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    return form.applAnchor(
      'read','View/Load',sid,
      [('dn',dn),('read_attr',self.attrType),('read_attrindex',str(valueindex))]
    )


class Audio(Binary):
  oid = '1.3.6.1.4.1.1466.115.121.1.4'
  desc = 'Audio'
  mimeType = 'audio/basic'
  fileExt = 'au'

  def _validate(self,attrValue):
    f = StringIO(attrValue)
    res = sndhdr.test_au(attrValue,f)
    return res!=None

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    mimetype = self.getMimeType(form)
    return """
      <embed
        type="%s"
        autostart="false"
        src="%s/read/%s?dn=%s&amp;read_attr=%s&amp;read_attrindex=%d"
      >
      %d bytes of audio data (%s)
      """ % (
        mimetype,
        form.script_name,sid,
        urllib.quote(utf2display(form.accept_charset,dn,escape_entities=0)),
        urllib.quote(self.attrType),
        valueindex,
        len(self.attrValue),
        mimetype
      )


class Certificate(Binary):
  oid = '1.3.6.1.4.1.1466.115.121.1.8'
  desc = 'Certificate Revocation List'
  mimeType = 'application/pkix-cert'
  fileExt = 'crt'

  def getMimeType(self,form):
    env = form.env
    if env.has_key('HTTP_USER_AGENT'):
      browsertype,browserversion = pyweblib.helper.BrowserType(
        env['HTTP_USER_AGENT']
      )
      if browsertype in ['Mozilla','Opera']:
        if self.attrType.lower().startswith('cacertificate'):
          return 'application/x-x509-ca-cert'
        else:
          return 'application/x-x509-email-cert'
    return self.mimeType

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    return form.applAnchor(
      'read','View/Load',sid,
      [
        ('dn',dn),
        ('read_attr',self.attrType),
        ('read_attrindex',str(valueindex)),
        ('read_attrmode','view'),
      ]
    )


class CertificateRevocationList(Certificate):
  oid = '1.3.6.1.4.1.1466.115.121.1.9'
  desc = 'Certificate Revocation List'
  mimeType = 'application/pkix-crl'
  fileExt = 'crl'

  def getMimeType(self,form):
    env = form.env
    if env.has_key('HTTP_USER_AGENT'):
      browsertype,browserversion = pyweblib.helper.BrowserType(
        env['HTTP_USER_AGENT']
      )
      if browsertype in ['Mozilla','Opera']:
        return 'application/x-pkcs7-crl'
    return self.mimeType


class DirectoryString(LDAPSyntax):
  oid = '1.3.6.1.4.1.1466.115.121.1.15'
  desc = 'Directory String'
  charset = 'utf-8'

  def _validate(self,attrValue):
    try:
      unicode(attrValue,self.charset).encode(self.charset)
    except UnicodeError:
      return 0
    else:
      return 1

  def sanitizeInput(self,attrValue):
    if type(attrValue)==StringType:
      return attrValue
    elif type(attrValue)==UnicodeType:
      return attrValue.encode(self.charset)
    else:
      raise ValueError, "Argument attrValue must be either StringType or UnicodeType"

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    return utf2display(form.accept_charset,self.attrValue)


class DistinguishedName(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.12'
  desc = 'Distinguished Name'

  def _validate(self,attrValue):
    return is_dn(attrValue)

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    return w2lapp.gui.DisplayDN(sid,form,ls,self.attrValue,commandbutton=commandbutton)


class NameAndOptionalUID(DistinguishedName):
  oid = '1.3.6.1.4.1.1466.115.121.1.34'
  desc = 'Name And Optional UID'

  def _splitDNandUID(self,v):
    try:
      sep_ind = attrValue.rindex('#')
    except ValueError:
      dn = v
      uid = None
    else:
      dn = attrValue[0:sep_ind]
      uid = attrValue[sep_ind+1:]
    return dn,uid

  def _validate(self,attrValue):
    dn,uid = self._splitDNandUID(attrValue)
    return is_dn(dn)

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    value = self.attrValue.split('#')
    dn_str = w2lapp.gui.DisplayDN(
      sid,form,ls,value[0],commandbutton=commandbutton
    )
    if len(value)==1 or not value[1]:
      return dn_str
    else:
      return ' | '.join([value[1],dn_str])
      

class Boolean(LDAPSyntax):
  oid = '1.3.6.1.4.1.1466.115.121.1.7'
  desc = 'Boolean'

  def _validate(self,attrValue):
    v = attrValue.upper()
    return (v=='TRUE') or (v=='FALSE')

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    return self.attrValue.encode('ascii')

  def formField(self,form):
    return pyweblib.forms.Select(
      self.attrType,
      ': '.join([self.attrType,self.desc]),1,
      options=('','TRUE','FALSE'),required=0,default=self.attrValue
    )


class CaseExactString(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.26'
  desc = 'Case Exact String'


class BitString(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.6'
  desc = 'Bit String'
  reObj=re.compile("'[01]+'B")


class GeneralizedTime(LDAPSyntax):
  oid = '1.3.6.1.4.1.1466.115.121.1.24'
  desc = 'Generalized Time'
  reObj=timestamp_regex

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    try:
      return utctime.strftimeiso8601(utctime.strptime(self.attrValue)).encode('ascii')
    except ValueError:
      return utf2display(form.accept_charset,self.attrValue)


class UTCTime(GeneralizedTime):
  oid = '1.3.6.1.4.1.1466.115.121.1.53'
  desc = 'UTC Time'


class IA5String(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.26'
  desc = 'IA5 String'
  charset = 'ascii'


class Integer(IA5String):
  oid = '1.3.6.1.4.1.1466.115.121.1.27'
  desc = 'Integer'

  def _validate(self,attrValue):
    try:
      int(attrValue)
    except ValueError:
      return 0
    else:
      return 1


class IntegerRange(Integer):
  oid = 'IntegerRange-oid'
  desc = 'Integer within a certain range'
  # The following class members are set to None to make direct use of this class fail
  minValue = None
  maxValue = None

  def _validate(self,attrValue):
    try:
      intValue = int(attrValue)
    except ValueError:
      return 0
    else:
      return (intValue>=self.minValue) and (intValue<=self.maxValue)

  def formField(self,form):
    max_len = max(len(str(self.minValue)),len(str(self.maxValue)))    
    return pyweblib.forms.Input(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      max_len,self.maxValues,'[0-9]*',default=self.attrValue,size=max_len
    )


class RFC822Address(IA5String):
  oid = 'RFC822Address-oid'
  desc = 'RFC 822 mail address'
  reObj= re.compile(mail_pattern)

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    return w2lapp.gui.MailToURL(self.attrValue)


class Uri(DirectoryString):
  oid = 'Uri-OID'
  desc = 'URI'
  reObj = url_regex

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    try:
      url,label = self.attrValue.split(' ',1)
    except ValueError:
      url,label = self.attrValue,self.attrValue
      display_url = ''
    else:
      display_url = ' (%s)' % (url)
    return '<a href="%s/urlredirect?%s">%s%s</a>' % (
      form.script_name,
      url.encode('ascii'),
      utf2display(form.accept_charset,label),
      utf2display(form.accept_charset,display_url),
    )


class JPEGImage(Binary):
  oid = '1.3.6.1.4.1.1466.115.121.1.28'
  desc = 'JPEG image'
  mimeType = 'image/jpeg'
  fileExt = 'jpg'

  def _validate(self,attrValue):
    return imghdr.test_jpeg(attrValue,None)=='jpeg'

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    return """
      <img
        src="%s/read/%s?dn=%s&amp;read_attr=%s&amp;read_attrindex=%d&amp;read_attrmode=load&amp;read_attrmimetype=image/jpeg"
        alt="%d bytes of image data"
      >
      """ % (
      form.script_name,sid,
      urllib.quote(utf2display(form.accept_charset,dn,escape_entities=0)),
      urllib.quote(self.attrType),
      valueindex,len(self.attrValue),
    )
#    return """
#      <img
#        src="data:image/jpeg;base64,%s"
#        alt="%d bytes of image data"
#      >
#      """ % (base64.encodestring(self.attrValue),len(self.attrValue))


class PhotoG3Fax(Binary):
  oid = '1.3.6.1.4.1.1466.115.121.1.23'
  desc = 'Photo (G3 fax)'
  mimeType = 'image/tiff'
  fileExt = 'tif'


class OID(IA5String):
  oid = '1.3.6.1.4.1.1466.115.121.1.38'
  desc = 'OID'
  regex = '([0-9]+.)*[0-9]'

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    attrValue = self.attrValue.encode('ascii')
    if oid_desc_reg.has_key(attrValue):
      name,description,reference = oid_desc_reg[self.attrValue]
      return '<strong>%s</strong> (%s):<br>%s (see %s)' % (
        name,attrValue,description,reference
      )
    else:
      return attrValue


class LDAPUrl(DirectoryString):
  oid = 'LDAPUrl-oid'
  desc = 'LDAP URL'

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    try:
      if commandbutton:
        commandbuttonstr = w2lapp.gui.LDAPURLButton(sid,form,ls,self.attrValue)
      else:
        commandbuttonstr = ''
      return '<table><tr><td>%s</td><td><a href="%s">%s</a></td></tr></table>' % (
               commandbuttonstr,
	       self.attrValue.encode(form.accept_charset),
	       utf2display(form.accept_charset,self.attrValue)
	     )
    except ValueError:
      return '<strong>Not a valid LDAP URL:</strong> %s' % utf2display(form.accept_charset,repr(self.attrValue))


class OctetString(Binary):
  oid = '1.3.6.1.4.1.1466.115.121.1.40'
  desc = 'Octet String'
  editable = 1

  def sanitizeInput(self,attrValue):
    for c in [':',' ',',','\r','\n']:
      attrValue = attrValue.replace(c,'')
    result = []
    for i in range(0,len(attrValue),2):
      result.append(chr(int(attrValue[i:i+2],16)))
    result_str = ''.join(result)
    return result_str

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    return '<code><font size="-1">%s</font></code>' % mspki.util.HexString(
      self.attrValue,
      delimiter=':',wrap=64,linesep='<br>'
    )

  def formField(self,form):
    attrValue_default = mspki.util.HexString(str(self.attrValue or ''),delimiter=':',wrap=40,linesep='\r\n')
    return pyweblib.forms.Textarea(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      10000,1,
      None,
      default=attrValue_default,
      rows=3,cols=41
    )


class PostalAddress(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.41'
  desc = 'Postal Address'
  reObj=re.compile('^.*$',re.S+re.M)
  def _split_lines(self,v):
    return v.split('$')

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    lines = [
      utf2display(form.accept_charset,l)
      for l in self._split_lines(self.attrValue)
    ]
    return '<br>'.join(lines)

  """
  def formField(self,form):
    return pyweblib.forms.Textarea(
      self.attrType,
      ': '.join([self.attrType,self.desc]),
      self.maxLen,self.maxValues,
      None,
      default=u'\r\n'.join(self._split_lines(self.attrValue)),
      rows=3,cols=40
    )
  """


class PrintableString(DirectoryString):
  oid = '1.3.6.1.4.1.1466.115.121.1.44'
  desc = 'Printable String'
  reObj= re.compile('^[a-zA-Z0-9"()+,.=/:? -]*$')
  charset = 'ascii'


class NumericString(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.36'
  desc = 'Numeric String'
  reObj= re.compile('^[0-9]+$')


class EnhancedGuide(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.21'
  desc = 'Enhanced Search Guide'


class Guide(EnhancedGuide):
  oid = '1.3.6.1.4.1.1466.115.121.1.25'
  desc = 'Search Guide'


class TelephoneNumber(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.50'
  desc = 'Telephone Number'
  reObj= re.compile('^\+[0-9 ]+$')


class FacsimileTelephoneNumber(TelephoneNumber):
  oid = '1.3.6.1.4.1.1466.115.121.1.22'
  desc = 'Facsimile Number'
  reObj= re.compile('^\+[0-9 ]+$')


class TelexNumber(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.52'
  desc = 'Telex Number'


class TeletexTerminalIdentifier(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.51'
  desc = 'Teletex Terminal Identifier'


class ObjectGUID(LDAPSyntax):
  oid = 'ObjectGUID-oid'
  desc = 'Object GUID'
  charset = 'ascii'

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    objectguid_str = ''.join(['%02X' % ord(c) for c in self.attrValue])
    return ldapurl.LDAPUrl(
      ldapUrl=ls.uri,
      dn='GUID=%s' % (objectguid_str),
      who=None,cred=None
    ).htmlHREF(
      hrefText=objectguid_str,
      hrefTarget=None
    )


class SecondsSinceEpoch(Integer):
  oid = 'SecondsSinceEpoch-oid'
  desc = 'Seconds since epoch (1970-01-01 00:00:00)'

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    try:
      return utctime.strftimeiso8601(time.gmtime(float(self.attrValue))).encode('ascii')
    except ValueError:
      return utf2display(form.accept_charset,self.attrValue)


class DaysSinceEpoch(Integer):
  oid = 'DaysSinceEpoch-oid'
  desc = 'Days since epoch (1970-01-01)'

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    try:
      return utctime.strftimeiso8601(time.gmtime(float(self.attrValue)*86400)).encode('ascii')
    except ValueError:
      return utf2display(form.accept_charset,self.attrValue)


class SelectList(LDAPSyntax):
  """
  Base class for dictionary based select lists which
  should not be used directly
  """

  attr_value_dict = {}
  
  def _sortedSelectOptions(self):
    tmplist = [(x[1], x) for x in self.attr_value_dict.items()]
    tmplist.sort()
    return [x for (k,x) in tmplist]
    
  def _validate(self,attrValue):
    return self.attr_value_dict.has_key(attrValue)

  def displayValue(self,sid,form,ls,dn,valueindex=0,commandbutton=0):
    attr_value_str = utf2display(form.accept_charset,self.attrValue)
    if self.attr_value_dict.has_key(self.attrValue):
      return ': '.join((
        attr_value_str,
        utf2display(form.accept_charset,self.attr_value_dict[self.attrValue])
      ))
    else:
      return attr_value_str

  def formField(self,form):
    return pyweblib.forms.Select(
      self.attrType,
      ': '.join([self.attrType,self.desc]),1,
      options=self._sortedSelectOptions(),required=0,default=self.attrValue
    )


class CountryString(SelectList):
  oid = '1.3.6.1.4.1.1466.115.121.1.11'
  desc = 'Two letter country string as listed in ISO 3166-2'
  attr_value_dict = ldap.cidict.cidict(w2lapp.cnf.countries.c_dict)


class DeliveryMethod(PrintableString):
  oid = '1.3.6.1.4.1.1466.115.121.1.14'
  desc = 'Delivery Method'
  pdm = '(any|mhs|physical|telex|teletex|g3fax|g4fax|ia5|videotex|telephone)'
  reObj= re.compile('^%s[ $]*%s$' % (pdm,pdm))


# Set up the central syntax registry instance
syntax_registry = SyntaxRegistry()
# Register all syntax classes in this module
for symbol_name in dir():
  syntax_registry.registerSyntaxClass(eval(symbol_name))

syntax_registry.registerAttrType(
  OID.oid,
  [
    '1.2.826.0.1050.11.0','ogSupportedProfile',
    '1.3.6.1.4.1.1466.101.120.13','supportedControl',
    '1.3.6.1.4.1.1466.101.120.7','supportedExtension',
    '1.3.6.1.4.1.4203.1.3.5','supportedFeatures',
    'supportedCapabilities',
  ]
)

syntax_registry.registerAttrType(
  RFC822Address.oid,
  [
    '0.9.2342.19200300.100.1.3',    # mail, rfc822Mailbox
    '2.16.840.1.113730.3.1.13',     # mailLocalAddress
    '2.16.840.1.113730.3.1.17',     # mailForwardingAddress
    '2.16.840.1.113730.3.1.30',     # mgrpRFC822MailMember
    '2.16.840.1.113730.3.1.47',     # mailRoutingAddress
    '1.2.840.113549.1.9.1',         # email, emailAddress, pkcs9email
    '2.16.840.1.113678.2.2.2.2.15', # Domino/LDAP: MailAddress
  ]
)

syntax_registry.registerAttrType(
  JPEGImage.oid,
  [
    '0.9.2342.19200300.100.1.60', # jpegPhoto
  ]
)

syntax_registry.registerAttrType(
  Audio.oid,
  [
    '0.9.2342.19200300.100.1.55', # audio
  ]
)

syntax_registry.registerAttrType(
  Uri.oid,[
    '1.3.6.1.4.1.250.1.57', # labeledURI
  ]
)

syntax_registry.registerAttrType(
  Boolean.oid,[
    '2.5.18.9', # hasSubordinates
  ]
)

syntax_registry.registerAttrType(
  PostalAddress.oid,[
    '2.5.4.16',                   # postalAddress
    '2.5.4.26',                   # registeredAddress
    '0.9.2342.19200300.100.1.39', # homePostalAddress
  ]
)

syntax_registry.registerAttrType(
  LDAPUrl.oid,[
    '2.16.840.1.113730.3.1.34', # ref
  ]
)

syntax_registry.registerAttrType(
  Certificate.oid,[
    '2.5.4.36', # userCertificate
    '2.5.4.37', # cACertificate
    'userCertificate','userCertificate;binary',
    'cACertificate','cACertificate;binary',
  ]
)

syntax_registry.registerAttrType(
  CertificateRevocationList.oid,[
    '2.5.4.38', # authorityRevocationList
    '2.5.4.39', # certificateRevocationList
    '2.5.4.53', # deltaRevocationList
    'authorityRevocationList','authorityRevocationList;binary',
    'certificateRevocationList','certificateRevocationList;binary',
    'deltaRevocationList','deltaRevocationList;binary',
  ]
)

syntax_registry.registerAttrType(
  UTCTime.oid,[
    '2.5.18.1', # createTimestamp
    '2.5.18.2', # modifyTimestamp
    'createtimestamp-oid', # createtimestamp on Netscape DS 4.x
    'modifytimestamp-oid', # modifytimestamp on Netscape DS 4.x
  ]
)

syntax_registry.registerAttrType(
  CountryString.oid,[
    'c',
    'countryName',
    '2.5.4.6', # c
  ]
)

# Some LDAP servers (e.g. MS AD) declare these attributes with OctetString
# syntax but Binary syntax is more suitable
syntax_registry.registerAttrType(
  Binary.oid,[
    '2.16.840.1.113730.3.1.216', # userPKCS12
    '2.16.840.1.113730.3.140',   # userSMIMECertificate
  ]
)

syntax_registry.registerAttrType(
  DistinguishedName.oid,[
    'namingContexts',
  ]
)

