"""
w2lapp.schema.viewer -  Display LDAPv3 schema

web2ldap - a web-based LDAP Client,
see http://www.web2ldap.de for details

(c) by Michael Stroeder <michael@stroeder.com>

This module is distributed under the terms of the
GPL (GNU GENERAL PUBLIC LICENSE) Version 2
(see http://www.gnu.org/copyleft/gpl.html)

$Id: viewer.py,v 1.63 2005/06/08 12:15:26 michael Exp $
"""

import types,ldap,w2lapp.gui

from w2lapp.core import utf2display

from ldap.schema import \
  SCHEMA_ATTRS,SCHEMA_CLASS_MAPPING,SCHEMA_ATTR_MAPPING, \
  LDAPSyntax,AttributeType,ObjectClass,MatchingRule,MatchingRuleUse,DITContentRule

from ldaputil.base import escape_filter_chars
from pyweblib.forms import escapeHTML

def LinkText(se):
  names = se.__dict__.get('names',(()))
  if len(names)==1:
    link_text = names[0]
  elif len(names)>1:
    link_text = '%s (alias %s)' % (names[0],', '.join(names[1:]))
  elif isinstance(se,LDAPSyntax) and not (se.desc is None):
    link_text = se.desc
  else:
    link_text = se.oid
  return link_text


def displayNameOrOID(sid,form,dn,schema,se_nameoroid,se_class):
  se = schema.get_obj(se_class,se_nameoroid,default=None)
  if se:
    link_text = LinkText(se)
    return form.applAnchor(
      'oid',link_text,sid,[
        ('dn',dn),('oid',se.oid),
        ('oid_class',SCHEMA_ATTR_MAPPING[se_class])
      ]
    )
  else:
    return se_nameoroid


def displayNameOrOIDList(sid,form,dn,schema,se_names,se_class):
  link_texts = []
  for se_nameoroid in se_names:
    se = schema.get_obj(se_class,se_nameoroid,default=None)
    if se:
      link_text = LinkText(se)
      anchor = form.applAnchor(
        'oid',link_text,sid,[
          ('dn',dn),('oid',se.oid),
          ('oid_class',SCHEMA_ATTR_MAPPING[se_class])
        ]
      )
      link_texts.append((link_text,anchor))
    else:
      link_texts.append((se_nameoroid,se_nameoroid))
  link_texts.sort()
  return [ i[1] for i in link_texts ]


def HTMLSchemaTree(sid,outf,form,dn,schema,se_class,se_tree,se_oid,level):
  """HTML output for browser"""
  outf.write('<dl>')
  se_obj = schema.get_obj(se_class,se_oid)
  if se_obj!=None:
    display_id = (se_obj.names or (se_oid,))[0]
    outf.write("""
    <dt><strong>%s</strong></dt>
    """ % (displayNameOrOID(sid,form,dn,schema,display_id,se_class)))
  if se_tree[se_oid]:
    outf.write('<dd>')
    for sub_se_oid in se_tree[se_oid]:
      HTMLSchemaTree(sid,outf,form,dn,schema,se_class,se_tree,sub_se_oid,level+1)
    outf.write('</dd>')
  outf.write('</dl>')


def SchemaContextMenu(sid,form,ls,dn):
  """Build context menu with schema-related items"""
  context_menu_list = []
  try:
    subschemaSubentryDN = ls.searchSubSchemaEntryDN(dn)
  except ldap.LDAPError:
    browse_link_text = 'Browse Local Schema'
  else:
    if subschemaSubentryDN!=None:
      form_param_list = [
        ('dn',subschemaSubentryDN),
      ]
      for schema_attr in SCHEMA_ATTRS+['objectClass','cn']:
        form_param_list.append(('read_attr',schema_attr))
      context_menu_list.append(form.applAnchor('read','Subschema Subentry',sid,form_param_list))
      browse_link_text = 'Browse Subschema'
    else:
      browse_link_text = 'Browse Local Schema'
  context_menu_list.append(form.applAnchor('oid',browse_link_text,sid,[('dn',dn)]))
  return context_menu_list


class DisplaySchemaElement:
  type_desc = 'Abstract Schema Element'
  detail_attrs = ()

  def __init__(self,sub_schema,se):
    self.s = sub_schema
    self.se = se
    self.sei = sub_schema.get_inheritedobj(self.se.__class__,self.se.oid,[])

  def displayDetails(self,sid,outf,form,dn):
    for text,class_attr,se_class in self.detail_attrs:
      class_attr_value = self.sei.__dict__.get(class_attr,None)
      if not class_attr_value:
        continue
      if type(class_attr_value)==types.TupleType or \
         type(class_attr_value)==types.ListType:
        class_attr_value_list = list(class_attr_value)
        class_attr_value_list.sort()
      else:
        class_attr_value_list = [class_attr_value]
      if se_class is None:
        value_output = ', '.join([w2lapp.core.utf2display(form.accept_charset,v) for v in class_attr_value_list])
      else:
        value_output = ', '.join(displayNameOrOIDList(sid,form,dn,self.s,class_attr_value_list,se_class))
      outf.write('<dt>%s</dt>\n<dd><p>\n%s\n</p></dd>\n' % (text,value_output))
    return # displayDetails()

  def display(self,sid,outf,form,ls,dn):
    w2lapp.gui.TopSection(
      sid,outf,form,ls,dn,
      '%s %s (%s)' % (
        self.type_desc,
        ', '.join(
          self.se.__dict__.get('names',(()))
        ),
        self.se.oid
      ),
      w2lapp.gui.MainMenu(sid,form,ls,dn),
      context_menu_list=SchemaContextMenu(sid,form,ls,dn)
    )
    outf.write("""<div id="Message" class="Main">
%s
<h1>%s <em>%s</em> (%s)</h1>
<a id="alvestrand_oid" href="%s/urlredirect?http://www.alvestrand.no/objectid/%s.html">
Try a lookup in Alvestrand's OID registry</a>
<dl>
<dt>Schema element string:</dt>
<dd><code>%s</code></dd>
    """ % (
      OIDInputForm(form,sid,dn,''),
      self.type_desc,
      ", ".join(
        self.se.__dict__.get('names',(()))
      ),
      self.se.oid,
      form.script_name,self.se.oid,
      utf2display(form.accept_charset,str(self.se))
    ))
    if ls.rootDSE.has_key('schemaNamingContext'):
      try:
        result = ls.l.search_s(
          ls.rootDSE['schemaNamingContext'][0],
          ldap.SCOPE_SUBTREE,
          '(|(&(objectClass=attributeSchema)(attributeID=%s))(&(objectClass=classSchema)(governsID=%s)))' % (self.se.oid,self.se.oid),
          attrlist=['cn']
        )
      except ldap.LDAPError:
        pass
      else:
        if result:
          dn,entry = result[0]
          outf.write('<dt>Schema Definition Entry (MS AD)</dt>\n<dd>\n%s\n</dd>\n' % (
            form.applAnchor(
              'read',entry['cn'][0],sid,
              [
                ('dn',dn),
              ],
          )))
    self.displayDetails(sid,outf,form,dn)
    outf.write('</dl></div>')
    w2lapp.gui.PrintFooter(outf,form)


class DisplayObjectClass(DisplaySchemaElement):
  type_desc = 'Object class'
  detail_attrs = (
    ('Description','desc',None),
    ('Derived from','sup',ObjectClass),
  )

  def __init__(self,sub_schema,se):
    DisplaySchemaElement.__init__(self,sub_schema,se)
    self.sei = sub_schema.get_inheritedobj(self.se.__class__,self.se.oid,['kind'])

  def displayDetails(self,sid,outf,form,dn):
    DisplaySchemaElement.displayDetails(self,sid,outf,form,dn)
    must,may = self.s.attribute_types([self.se.oid],raise_keyerror=0)
    # Display all required and allowed attributes
    outf.write('<dt>Kind of object class:</dt><dd>\n%s&nbsp;</dd>\n' % (
      {0:' STRUCTURAL',1:' ABSTRACT',2:' AUXILIARY'}[self.sei.kind]
    ))
    # Display all required and allowed attributes
    outf.write('<dt>All required attributes:</dt><dd>\n%s&nbsp;</dd>\n' % (
      ', '.join(displayNameOrOIDList(sid,form,dn,self.s,must.keys(),AttributeType))
    ))
    outf.write('<dt>All allowed attributes:</dt><dd>\n%s&nbsp;</dd>\n' % (
      ', '.join(displayNameOrOIDList(sid,form,dn,self.s,may.keys(),AttributeType))
    ))

    # Display relationship to DIT content rule(s)
    if self.sei.kind==0:
      # STRUCTURAL object class
      content_rule = self.s.get_obj(DITContentRule,self.se.oid)
      if content_rule:
        outf.write('<dt>Governed by DIT content rule:</dt><dd>\n%s&nbsp;</dd>\n' % (
          displayNameOrOID(sid,form,dn,self.s,content_rule.oid,DITContentRule)
        ))
        outf.write('<dt>Applicable auxiliary object classes:</dt><dd>\n%s&nbsp;</dd>\n' % (
          ', '.join(displayNameOrOIDList(sid,form,dn,self.s,content_rule.aux,ObjectClass))
        ))
    elif self.sei.kind==2:
      # AUXILIARY object class
      structural_oc_list = []
      for content_rule_name in self.s.listall(DITContentRule):
        content_rule = self.s.get_obj(DITContentRule,content_rule_name)
        for aux_class_name in content_rule.aux:
          aux_class_oid = self.s.getoid(ObjectClass,aux_class_name)
          if aux_class_oid==self.se.oid:
            structural_oc_list.append(content_rule.oid)
      if structural_oc_list:
        outf.write('<dt>Allowed with structural object classes:</dt><dd>\n%s&nbsp;</dd>\n' % (
          ', '.join(displayNameOrOIDList(sid,form,dn,self.s,structural_oc_list,ObjectClass))
        ))
    
    # Display tree of derived object classes
    outf.write('<dt>Object class tree:</dt>\n')
    outf.write('<dd>\n')
    try:
      oc_tree = self.s.tree(ObjectClass)
    except KeyError,e:
      outf.write('<strong>Missing schema elements referenced:<pre>%s</pre></strong>\n' % utf2display(form.accept_charset,str(e)))
    else:
      if oc_tree.has_key(self.se.oid) and oc_tree[self.se.oid]:
        HTMLSchemaTree(sid,outf,form,dn,self.s,ObjectClass,oc_tree,self.se.oid,0)
    outf.write('&nbsp;</dd>\n')
    # Display a link for searching entries by object class
    search_filter = '(objectClass=%s)' % (escape_filter_chars((self.se.names or [self.se.oid])[0]))
    outf.write('<dt>Search entries</dt>\n<dd>\n%s\n</dd>\n' % (
      form.applAnchor(
	'searchform',search_filter,sid,
        [
          ('dn',dn),
          ('searchform_mode','exp'),
          ('search_filterstr',search_filter),
        ],
        title='Search entries by object class',
    )))
    

class DisplayAttributeType(DisplaySchemaElement):
  type_desc = 'Attribute type'
  detail_attrs = (
    ('Description','desc',None),
    ('Syntax','syntax',LDAPSyntax),
    ('Derived from','sup',AttributeType),
    ('Equality matching rule','equality',MatchingRule),
    ('Sub-string matching rule','substr',MatchingRule),
    ('Ordering matching rule','ordering',MatchingRule),
  )

  def __init__(self,sub_schema,se):
    DisplaySchemaElement.__init__(self,sub_schema,se)
    self.sei = sub_schema.get_inheritedobj(self.se.__class__,self.se.oid,['syntax','equality','substr','ordering'])

  def displayDetails(self,sid,outf,form,dn):

    DisplaySchemaElement.displayDetails(self,sid,outf,form,dn)

    at_oid = self.se.oid
    syntax_oid = self.sei.syntax

    if syntax_oid!=None:

      ####################################
      # Display applicable matching rules
      ####################################

      mr_use_se = self.s.get_obj(MatchingRuleUse,syntax_oid)

      applies_dict = {}
      for mr_oid in self.s.listall(MatchingRuleUse):
        applies_dict[mr_oid] = {}
        mr_use_se = self.s.get_obj(MatchingRuleUse,mr_oid)
        for a in mr_use_se.applies:
          applies_dict[mr_oid][self.s.getoid(AttributeType,a)] = None
      # Display list of attribute types for which this matching rule is applicable
      mr_applicable_for = [
        mr_oid
        for mr_oid in self.s.listall(MatchingRule)
        if applies_dict.has_key(mr_oid) and applies_dict[mr_oid].has_key(at_oid)
      ]
      if mr_applicable_for:
        outf.write('<dt>Applicable matching rules:</dt>\n<dd>\n%s\n</dd>\n' % (
          ', '.join(displayNameOrOIDList(sid,form,dn,self.s,mr_applicable_for,MatchingRule))
        ))

    ###################################################################
    # Display object classes which may contain attributes of this type
    ###################################################################
    all_object_classes = self.s.listall(ObjectClass)
    attr_type_usable_in = []
    for oc_oid in all_object_classes:
      must,may = self.s.attribute_types([oc_oid],raise_keyerror=0)
      if must.has_key(at_oid) or may.has_key(at_oid):
        attr_type_usable_in.append(oc_oid)
    if attr_type_usable_in:
      outf.write('<dt>Attribute type usable in:</dt>\n<dd>\n%s\n</dd>\n' % (
        ', '.join(displayNameOrOIDList(sid,form,dn,self.s,attr_type_usable_in,ObjectClass))
      ))

    #########################################
    # Output attribute type inheritance tree
    #########################################
    outf.write('<dt>Attribute type tree:</dt>\n')
    outf.write('<dd>\n')
    # Display tree of derived attribute types
    try:
      at_tree = self.s.tree(AttributeType)
    except KeyError,e:
      outf.write('<strong>Missing schema elements referenced:<pre>%s</pre></strong>\n' % utf2display(form.accept_charset,str(e)))
    else:
      if at_tree.has_key(at_oid) and at_tree[at_oid]:
        HTMLSchemaTree(sid,outf,form,dn,self.s,AttributeType,at_tree,at_oid,0)
    outf.write('</dd>\n')
    # Display a link for searching entries by attribute presence
    search_filter = '(%s=*)' % (escape_filter_chars((self.se.names or [self.se.oid])[0]))
    outf.write('<dt>Search entries</dt>\n<dd>\n%s\n</dd>\n' % (
      form.applAnchor(
	'searchform',search_filter,sid,
        [
          ('dn',dn),
          ('searchform_mode','exp'),
          ('search_filterstr',search_filter),
        ],
        title='Search entries by attribute presence',
    )))
    

class DisplayLDAPSyntax(DisplaySchemaElement):
  type_desc = 'LDAP Syntax'
  detail_attrs = (
    ('Description','desc',None),
  )

  def displayDetails(self,sid,outf,form,dn):
    DisplaySchemaElement.displayDetails(self,sid,outf,form,dn)
    # Display list of attribute types which directly reference this syntax
    syntax_used_for = [
      at_oid
      for at_oid in self.s.listall(AttributeType)
      if self.s.get_obj(AttributeType,at_oid).__class__==AttributeType and \
         self.s.get_syntax(at_oid)==self.se.oid
    ]
    if syntax_used_for:
      outf.write('<dt>Syntax used for:</dt>\n<dd>\n%s\n</dd>\n' % (
        ', '.join(displayNameOrOIDList(sid,form,dn,self.s,syntax_used_for,AttributeType))
      ))


class DisplayMatchingRule(DisplaySchemaElement):
  type_desc = 'Matching Rule'
  detail_attrs = (
    ('Description','desc',None),
    ('LDAP syntax','syntax',LDAPSyntax),
  )

  def displayDetails(self,sid,outf,form,dn):
    DisplaySchemaElement.displayDetails(self,sid,outf,form,dn)
    mr_use_se = self.s.get_obj(MatchingRuleUse,self.se.oid)
    if mr_use_se:
      applies_dict = {}
      for a in mr_use_se.applies:
        applies_dict[self.s.getoid(AttributeType,a)] = None
      # Display list of attribute types for which this matching rule is applicable
      mr_applicable_for = [
        at_oid
        for at_oid in self.s.listall(AttributeType)
        if applies_dict.has_key(at_oid)
      ]
      if mr_applicable_for:
        outf.write('<dt>Applicable for attribute types:</dt>\n<dd>\n%s\n</dd>\n' % (
          ', '.join(displayNameOrOIDList(sid,form,dn,self.s,mr_applicable_for,AttributeType))
        ))
    
    mr_used_by = []
    for at_oid in self.s.listall(AttributeType):
      at_se = self.s.get_inheritedobj(AttributeType,self.se.oid,['equality','substr','ordering'])
      if at_se and \
         (at_se.equality==self.se.oid or at_se.substr==self.se.oid or at_se.ordering==self.se.oid):
        mr_used_by.append(at_se.oid)
    if mr_used_by:
      outf.write('<dt>Used by attribute types:</dt>\n<dd>\n%s\n</dd>\n' % (
        ', '.join(displayNameOrOIDList(sid,form,dn,self.s,mr_used_by,AttributeType))
      ))



class DisplayMatchingRuleUse(DisplaySchemaElement):
  type_desc = 'Matching Rule Use'
  detail_attrs = (
    ('Names','names',None),
    ('Matching Rule','oid',MatchingRule),
    ('Applies to','applies',AttributeType),
  )


class DisplayDITContentRule(DisplaySchemaElement):
  type_desc = 'DIT content rule'
  detail_attrs = (
    ('Names','names',None),
    ('Governs structural object class','oid',ObjectClass),
    ('Auxiliary classes','aux',ObjectClass),
    ('Must have','must',AttributeType),
    ('May have','may',AttributeType),
    ('Must not have','nots',AttributeType),
  )


SCHEMA_VIEWER_CLASS = {
  ObjectClass:DisplayObjectClass,
  AttributeType:DisplayAttributeType,
  LDAPSyntax:DisplayLDAPSyntax,
  MatchingRule:DisplayMatchingRule,
  MatchingRuleUse:DisplayMatchingRuleUse,
  DITContentRule:DisplayDITContentRule,
}


def OIDInputForm(form,sid,dn,oid=''):
  return form.formHTML(
    'oid','Search',sid,'GET',
    [('dn',dn)],
    extrastr=w2lapp.form.OIDInput('oid',u'OID').inputHTML(oid)
  )


def DisplayAllSchemaelements(sid,outf,form,ls,dn,schema,se_list=None):
  se_list = se_list or []
  w2lapp.gui.TopSection(
    sid,outf,form,ls,dn,'Schema elements',
    w2lapp.gui.MainMenu(sid,form,ls,dn),
    context_menu_list=SchemaContextMenu(sid,form,ls,dn)
  )
  oid_dict = {}
  if se_list:
    for schema_class in SCHEMA_VIEWER_CLASS.keys():
      oid_dict[schema_class] = []
    for se in se_list:
      oid_dict[se.__class__].append(se.oid)
  else:
    for schema_class in SCHEMA_VIEWER_CLASS.keys():
      oid_dict[schema_class] = schema.listall(schema_class)
  outf.write('<div id="Message" class="Main">\n<h1>Schema elements</h1>\n')
  outf.write(OIDInputForm(form,sid,dn,''))
  for schema_class in SCHEMA_VIEWER_CLASS.keys():
    schema_elements = oid_dict[schema_class]
    if not schema_elements:
      continue
    outf.write('<h2>%s</h2>\n' % (SCHEMA_VIEWER_CLASS[schema_class].type_desc))
    outf.write('<p>found %d</p>\n' % (len(schema_elements)))
    outf.write(
      ', '.join(
        displayNameOrOIDList(
          sid,form,dn,schema,schema_elements,schema_class
        )
      )
    )    
  outf.write('</div>\n')
  w2lapp.gui.PrintFooter(outf,form)


def w2l_DisplaySchemaElement(sid,outf,command,form,ls,dn):

  sub_schema = ls.retrieveSubSchema(dn,w2lapp.cnf.GetParam(ls,'_schema',None))

  # Get input parameter from form input
  oid = form.getInputValue('oid',[None])[0]

  if oid is None:
    # Display entry page of schema browser
    DisplayAllSchemaelements(sid,outf,form,ls,dn,sub_schema)

  else:

    se_class = form.getInputValue('oid_class',[None])[0]
    
    if se_class:
      # Display a single schema element referenced by OID and class
      se_obj = sub_schema.get_obj(SCHEMA_CLASS_MAPPING[se_class],oid,None)
    else:
      # Search schema element by OID
      se_list = []
      for schema_element_type in SCHEMA_VIEWER_CLASS.keys():
        se = sub_schema.get_obj(schema_element_type,oid,None)
        if not se is None:
          se_list.append(se)
      if not se_list:
        raise w2lapp.core.ErrorExit('Name or OID not found in schema!')
      elif len(se_list)>1:
        DisplayAllSchemaelements(sid,outf,form,ls,dn,sub_schema,se_list)
        return
      se_obj = se_list[0]

    if not SCHEMA_VIEWER_CLASS.has_key(se_obj.__class__):
      raise w2lapp.core.ErrorExit('No viewer for this type of schema element!')
    schema_viewer = SCHEMA_VIEWER_CLASS[se_obj.__class__](sub_schema,se_obj)
    schema_viewer.display(sid,outf,form,ls,dn)
